from ._mainOptic import *
from .Absorber import *
from .Laser import *
from .OpticEnums import *

# Docstring for the AOT_Optic package
"""
AOT_Optic is a package for optical components in Acousto-Optic Tomography.
It provides tools and classes for working with optical elements such as absorbers and lasers.
"""